%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
%}

%union {
  double d;
  char* s;
  struct expr* expr;
  struct print* print;
  struct varlist* varlist;
  struct inst* inst;
}

%{

  typedef struct stack {
    double val;
    struct stack* next;
  } stack;

  typedef struct expr {
    int    type;
    
    char*  var;         /* for VARIABLE type */
    double d;           /* for NOMBRE type */

    struct expr* left;  /* for unary and binary nodes */
    struct expr* right; /* for binary nodes */
    struct expr* next;  /* for expression lists */
  } expr;
  
  typedef struct print { /* arguments of a PRINT construct */
    int type;
    
    char* s;             /* if type is CHAINE */
    expr* expr;   /* if type is NOMBRE */
    
    struct print* next;  /* next object */
  } print;
  
  typedef struct varlist {
    char* var;
    stack* stack;
    struct varlist* next;
  } varlist;
  
  typedef struct inst {
    int type;
    
    expr* expr1;         /* test in IF, WHILE, 1st expr in FOR, call */
    expr* expr2;         /* 2d expr in FOR */
    expr* expr3;         /* 3d expr in FOR */
    print* print;        /* PRINT object */
    struct inst*  block; /* block in IF, WHILE and FOR */
    char* s;             /* string for INPUT */
    char* var;           /* variable in EGAL, FOR and INPUT, procedure */
    varlist* varlist;    /* for procedure definition */
    
    struct inst* next;   /* next instruction if in a bloc */
    
  } inst;


  expr* expr_bin(int type, expr* e1, expr* e2){
    expr* e = (expr*)malloc(sizeof(expr));
    e->type=type; e->left=e1; e->right=e2;
    return e;
  }

  expr* expr_nb(double d) {
    expr* e = (expr*)malloc(sizeof(expr));
    e->type=NOMBRE; e->d=d;
    return e;
  }

  expr* expr_var(char* var) {
    expr* e = (expr*)malloc(sizeof(expr));
    e->type=VARIABLE; e->var=var;
    return e;
  }

  inst* inst_while(expr* e1, inst* block) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=WHILE; i->expr1=e1; i->block=block;
    return i;
  }

  inst* inst_if(expr* e1, inst* block) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=IF; i->expr1=e1; i->block=block;
    return i;
  }

  inst* inst_for(char* var, expr* e1, expr* e2, 
		 expr* e3, inst* block) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=FOR; 
    i->var=var; i->expr1=e1; i->expr2=e2; i->expr3=e3; i->block=block;
    return i;
  }

  inst* inst_print(print* print) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=PRINT; i->print=print;
    return i;
  }

  inst* inst_input(char* s, char* var) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=INPUT; i->s=s; i->var=var;
    return i;
  }

  inst* inst_assign(char* var, expr* e1) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=EGAL; i->var=var; i->expr1=e1;
    return i;
  }

  inst* inst_simple(int type, expr* e1, expr* e2, expr* e3) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=type; i->expr1=e1; i->expr2=e2; i->expr3=e3;
    return i;
  }

  inst* inst_proc(char* var, varlist* arg, inst* block) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=PROC; i->var=var; i->varlist=arg; i->block=block;
    return i;
  }

  inst* inst_call(char* var, expr* arg) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=VARIABLE; i->var=var; i->expr1=arg;
    return i;
  }

  varlist* varlist_var(char* var) {
    varlist* v = (varlist*)malloc(sizeof(varlist));
    v->stack = NULL;
    v->var = var;
    return v;
  }

  print* print_chaine(char* s) {
    print* p = (print*)malloc(sizeof(print));
    p->type=CHAINE; p->s=s;
    return p;
  }

  print* print_expr(expr* expr) {
    print* p = (print*)malloc(sizeof(print));
    p->type=NOMBRE; p->expr=expr;
    return p;
  }


  typedef struct var {
    char* name;
    double val;
    struct var* next;
  } var;
  var vars = { "", 0, NULL };
 
  var* get_var(char *s) {
    var* v;
    for (v=&vars;v;v=v->next) 
      if (!strcmp(v->name,s)) return v;
    for (v=&vars;v->next;v=v->next);
    v->next=(var*)malloc(sizeof(var));
    v->next->name = s;
    v->next->next = NULL;
    return v->next;
  }

  double get_val(char* s) {
    var* v;
    for (v=&vars;v;v=v->next) if (!strcmp(v->name,s)) return v->val;
    return 0;
  }

  typedef struct proc {
    char* name;
    varlist* args;
    inst* block;
    struct proc* next;
  } proc;
  proc procs = {"",NULL,NULL,NULL};

  FILE* graph;
  double x=594/2,y=838/2,ang=0;
  int down=1;

  double eval_expr(expr* e) {
    switch (e->type) {
    case NOMBRE: return e->d;
    case PLUS: return eval_expr(e->left)+eval_expr(e->right);
    case MOINS: return eval_expr(e->left)-eval_expr(e->right);
    case MULTIPLIE: return eval_expr(e->left)*eval_expr(e->right);
    case DIVISE: return eval_expr(e->left)/eval_expr(e->right);
    case MODULO: return fmod(eval_expr(e->left),eval_expr(e->right));
    case OPPOSE: return -eval_expr(e->left);
    case RANDOM: return lrand48() % ((int)eval_expr(e->left));
    case ABS: return fabs(eval_expr(e->left));
    case EGAL: return (eval_expr(e->left)==eval_expr(e->right) ? 1 : 0);
    case DIFFERENT: return (eval_expr(e->left)!=eval_expr(e->right) ? 1 : 0);
    case SUPERIEUR: return (eval_expr(e->left)>eval_expr(e->right) ? 1 : 0);
    case SUPERIEUR_EGAL: return (eval_expr(e->left)>=eval_expr(e->right) ? 1 : 0);
    case INFERIEUR: return (eval_expr(e->left)<eval_expr(e->right) ? 1 : 0);
    case INFERIEUR_EGAL: return (eval_expr(e->left)<=eval_expr(e->right) ? 1 : 0);
    case VARIABLE: return get_val(e->var);

    default: 
      printf("\nInternal error: case %i in exec_expr\n\n",e->type); 
      exit(1);
    }
  }

  void exec_block(inst* i) {
    while (i) {
      switch (i->type) {
	
      case EGAL:
	get_var(i->var)->val = eval_expr(i->expr1);
	break;

      case INPUT: {
	float d;
	printf("%s",i->s);
	scanf("%e",&d);
	get_var(i->var)->val = d;
	break;
      }

      case PRINT: {
	print* p;
	for (p=i->print;p;p=p->next)
	  if (p->type==NOMBRE) printf("%g",eval_expr(p->expr));
	  else printf ("%s",p->s);
	break;
      }

      case WHILE: 
	while (eval_expr(i->expr1)==1) exec_block(i->block);
	break;

      case FOR: {
	double e1 = eval_expr(i->expr1);
	double e2 = eval_expr(i->expr2);
	double e3 = i->expr3 ? eval_expr(i->expr3) : 1;
	var* v = get_var(i->var);
	if (e1<=e2) 
	  for (v->val=e1;v->val<=e2;v->val+=e3) exec_block(i->block);
	else
	  for (v->val=e2;v->val>=e1;v->val+=e3) exec_block(i->block);	
	break;
      }

      case IF:
	if (eval_expr(i->expr1)==1) exec_block(i->block);
	break;

      case PROC: {
	proc* p;
	for (p=&procs;p->next;p=p->next);
	p->next=(proc*)malloc(sizeof(proc));
	p->next->name = i->var;
	p->next->block = i->block;
	p->next->args = i->varlist;
	p->next->next = NULL;
	break;
      }

      case VARIABLE: {
 	proc* p;
	varlist* l;
	expr* e;
	for (p=&procs;p;p=p->next)
	  if (!strcmp(p->name,i->var)) break;
	if (!p) { printf("Procedure %s non definie!\n",i->var); exit(1); }

	for (e=i->expr1,l=p->args;e || l;e=e->next,l=l->next)
	  if (!l || !e) { printf("Mauvais nombre d'arguments dans l'appel de procedure %s!\n",i->var); exit(1); }
	  else {
	    stack *s = (stack*)malloc(sizeof(stack));
	    s->next = l->stack;
	    s->val = eval_expr(e);
	    l->stack = s;
	  }

	for (l=p->args;l;l=l->next) {
	  double d;
	  stack* s = l->stack;
	  var* v = get_var(l->var);
	  d = v->val; v->val=s->val; s->val = d;
	}

	exec_block(p->block);
	
	for (l=p->args;l;l=l->next) {
	  stack* s = l->stack;
	  var* v = get_var(l->var);
	  v->val = s->val;
	  l->stack = s->next;
	  free(s);
	}
	break;
      }

      case TD: ang-=eval_expr(i->expr1); break;
      case TG: ang+=eval_expr(i->expr1); break;	

      case AV: {
	double a = eval_expr(i->expr1);
	double xx = x+a*cos(ang/180.*3.1415927);
	double yy = y+a*sin(ang/180.*3.1415927);
	if (down) fprintf(graph,"%f %f moveto %f %f lineto stroke\n",x,y,xx,yy);
	x=xx; y=yy;
	break;
      }

      case RE: {
	double a = eval_expr(i->expr1);
	double xx = x-a*cos(ang/180.*3.1415927);
	double yy = y-a*sin(ang/180.*3.1415927);
	if (down) fprintf(graph,"%f %f moveto %f %f lineto stroke\n",x,y,xx,yy);
	x=xx; y=yy;
	break;
      }

      case DOWN: down=1; break;
      case UP:   down=0; break;

      case COLOR:
	fprintf(graph,"%f %f %f setrgbcolor\n",
		eval_expr(i->expr1),eval_expr(i->expr2),eval_expr(i->expr3));
	break;

      default: 
	printf("\nInternal error: case %i in exec_block\n\n",i->type); 
	exit(1);
      }
      i=i->next;
    } 
  }
  
  void prog(inst* i) {
    graph = fopen("out.ps","w");
    fprintf(graph,"%%!postscript\n");
    exec_block(i);
    printf("\n");
    fprintf(graph,"showpage\n");
    fclose(graph);
  }

%}

%token PRINT INPUT IF FOR TO STEP PROC WHILE
%token PLUS MOINS MULTIPLIE DIVISE MODULO RANDOM ABS
%token SUPERIEUR SUPERIEUR_EGAL INFERIEUR INFERIEUR_EGAL DIFFERENT EGAL
%token PARENTHESE_GAUCHE PARENTHESE_DROITE VIRGULE DEUXPOINTS
%token BLOC_GAUCHE BLOC_DROITE
%token ERROR QUIT
%token NOMBRE CHAINE VARIABLE
%token AV TG TD RE UP DOWN COLOR

%left SUPERIEUR SUPERIEUR_EGAL INFERIEUR INFERIEUR_EGAL DIFFERENT EGAL
%left PLUS MOINS
%left MULTIPLIE DIVISE MODULO
%nonassoc OPPOSE 

%start prog

%type<inst> instrliste instr block
%type<expr> expr exprlist
%type<print> printexprlist printexpr
%type<d> NOMBRE
%type<s> VARIABLE CHAINE
%type<varlist> varlist

%%

prog:   instrliste QUIT { prog($1); YYACCEPT; }

instrliste:  instr { $$=$1; $$->next=NULL; }
| instr DEUXPOINTS instrliste { $$=$1; $$->next=$3; }

instr: INPUT CHAINE VIRGULE VARIABLE { $$=inst_input($2,$4); }
| PRINT printexprlist { $$=inst_print($2); }
| IF expr block { $$=inst_if($2,$3); }
| VARIABLE EGAL expr { $$=inst_assign($1,$3); }
| FOR VARIABLE EGAL expr TO expr block { $$=inst_for($2,$4,$6,NULL,$7); }
| FOR VARIABLE EGAL expr TO expr STEP expr block { $$=inst_for($2,$4,$6,$8,$9); }
| WHILE expr block { $$=inst_while($2,$3); }
| PROC VARIABLE PARENTHESE_GAUCHE varlist PARENTHESE_DROITE block { $$=inst_proc($2,$4,$6); }
| VARIABLE PARENTHESE_GAUCHE exprlist PARENTHESE_DROITE { $$=inst_call($1,$3); }
| AV expr { $$=inst_simple(AV,$2,NULL,NULL); }
| RE expr { $$=inst_simple(RE,$2,NULL,NULL); }
| TD expr { $$=inst_simple(TD,$2,NULL,NULL); }
| TG expr { $$=inst_simple(TG,$2,NULL,NULL); }
| UP      { $$=inst_simple(UP,NULL,NULL,NULL); }
| DOWN    { $$=inst_simple(DOWN,NULL,NULL,NULL); }
| COLOR expr VIRGULE expr VIRGULE expr { $$=inst_simple(COLOR,$2,$4,$6); }

block:  BLOC_GAUCHE instrliste BLOC_DROITE { $$=$2; }

expr:  VARIABLE { $$=expr_var($1); }
| expr PLUS expr { $$=expr_bin(PLUS,$1,$3); }
| expr MOINS expr  { $$=expr_bin(MOINS,$1,$3); }
| expr MULTIPLIE expr  { $$=expr_bin(MULTIPLIE,$1,$3); }
| expr DIVISE expr  { $$=expr_bin(DIVISE,$1,$3); }
| expr MODULO expr { $$=expr_bin(MODULO,$1,$3); }
| expr SUPERIEUR expr  { $$=expr_bin(SUPERIEUR,$1,$3); }
| expr SUPERIEUR_EGAL expr  { $$=expr_bin(SUPERIEUR_EGAL,$1,$3); }
| expr INFERIEUR expr  { $$=expr_bin(INFERIEUR,$1,$3); }
| expr INFERIEUR_EGAL expr  { $$=expr_bin(INFERIEUR_EGAL,$1,$3); }
| expr DIFFERENT expr  { $$=expr_bin(DIFFERENT,$1,$3); }
| expr EGAL expr  { $$=expr_bin(EGAL,$1,$3); }
| MOINS %prec OPPOSE expr  { $$=expr_bin(OPPOSE,$2,NULL); }
| PARENTHESE_GAUCHE expr PARENTHESE_DROITE { $$=$2; }
| RANDOM PARENTHESE_GAUCHE expr PARENTHESE_DROITE { $$=expr_bin(RANDOM,$3,NULL); }
| ABS PARENTHESE_GAUCHE expr PARENTHESE_DROITE { $$=expr_bin(ABS,$3,NULL); }
| NOMBRE { $$=expr_nb($1); }

exprlist: expr { $$=$1; $$->next=NULL; }
| expr VIRGULE exprlist { $$=$1; $$->next=$3; }

varlist: VARIABLE { $$=varlist_var($1); $$->next=NULL; }
| VARIABLE VIRGULE varlist { $$=varlist_var($1); $$->next=$3; }

printexprlist:  printexpr  { $$=$1; $$->next=NULL; }
| printexpr VIRGULE printexprlist { $$=$1; $$->next=$3; }

 printexpr: CHAINE { $$=print_chaine($1); }
 | expr { $$=print_expr($1); }
    
%%

int main (int argc, char** argv)
{
  if (argc==2) {
    FILE* f = fopen(argv[1],"r");
    if (f!=NULL) {
      yy_switch_to_buffer(yy_create_buffer(f,4096));
      yyparse();
      fclose(f);
      return 0;
    }
    printf("Fichier %s introuvavle!\n",argv[1]);
  }
  printf("interpreteur programme\n");
  return 1;
}
